#include "Table.h"
#include "IoTString.h"
#include "TimingSingleton.h"
#include "TransactionStatus.h"

#define NUMBER_OF_SENSORS 1
#define MACHINE_ID 15512

int main(int numargs, char ** args) {
	TimingSingleton * timer = TimingSingleton_getInstance();

	bool foundError = false;
	Vector<TransactionStatus *> * transStatusList = new Vector<TransactionStatus *>();

	// Setup the 2 clients
	IoTString * baseurl = new IoTString("http://dc-6.calit2.uci.edu/test.iotcloud/");
	IoTString * password = new IoTString("reallysecret");
	Table * t1 = new Table(baseurl, password, MACHINE_ID, -1);
	t1->rebuild();
	printf("T1 Ready\n");

	baseurl->releaseRef();
	password->releaseRef();

	char keyBuffer[80];
	char dataBuffer[80];

	sprintf(keyBuffer, "tempF0");
	IoTString * iKeyA = new IoTString(keyBuffer);
	sprintf(dataBuffer, "data4");
	IoTString * iValueA = new IoTString(dataBuffer);
	t1->createNewKey(iKeyA, MACHINE_ID);
	t1->startTransaction();
	t1->put(iKeyA, iValueA);
	transStatusList->add(t1->commitTransaction());
	
	sprintf(keyBuffer, "humid0");
	IoTString * iKeyB = new IoTString(keyBuffer);
	sprintf(dataBuffer, "data6");
	IoTString * iValueB = new IoTString(dataBuffer);
	t1->createNewKey(iKeyB, MACHINE_ID);
	t1->startTransaction();
	t1->put(iKeyB, iValueB);
	transStatusList->add(t1->commitTransaction());
	t1->update();
	
	iKeyB->releaseRef();
	iValueB->releaseRef();
	iKeyA->releaseRef();
	iValueA->releaseRef();

	for (uint i = 0; i < transStatusList->size(); i++) {
		TransactionStatus * status = transStatusList->get(i);
		if (status->getStatus() != TransactionStatus_StatusCommitted) {
			foundError = true;
			printf("Status error\n");
		}
		delete status;
	}
	
	if (foundError) {
		printf("Found Errors...\n");
	} else {
		printf("No Errors Found...\n");
	}

	delete transStatusList;
	delete t1;
}

